class Constants:

    VERSION = "3.0.11"

    # encoding constants
    FEATURE_DELIMITER = "-"

    # importing constants
    UNKNOWN = ''
    ALLELE_DELIMITER = "*"
    GENE_DELIMITER = "-"
    STOP_CODON = "*"
    CACHE_TYPE = "cache_type"
    CACHE_PATH = 'cache_path'
    COMMENT_SIGN = "#"
    NOT_COMPUTED = "not computed"
    SIGNAL_DELIMITER = "__"
    GAP_LETTER = "-"

    # plotting constants
    PLOTLY_BLACK = "#2A3F5E"
