from enum import Enum


class SymbolType(Enum):

    MOTIF = 0
    SIGNAL = 1
    ML_METHOD = 2
    DATASET = 3
    ENCODING = 4
    REPORT = 5
    SIMULATION = 6
    PREPROCESSING = 7
    INSTRUCTION = 8
    OUTPUT = 9
    WEIGHTING = 10
