<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>immuneML: Dataset Export Overview</title>
    <style>
        {{{css_style}}}
    </style>
    <meta name="viewport" content="width=device-width, initial-scale=1">
</head>
<body>
    <div class="container">
        <div>
            <h1>Dataset export: {{name}}</h1>
            <p>immuneML version: {{immuneML_version}}</p>
            <p>Full specification is available <a href="{{{full_specs}}}">here</a>.</p>
            <p>Log file is available <a href="{{{logfile}}}">here</a>.</p>
            <p></p>

            {{#datasets}}
            <div class="padded">
                <div class="col-container">
                    <div class="col">
                        <h3>Dataset information</h3>
                        <h4>General information</h4>
                        <div class="table-container">
                            <table>
                            <tr>
                                <td><b>Name</b></td>
                                <td>{{dataset_name}}</td>
                            </tr>
                            <tr>
                                <td><b>Type</b></td>
                                <td>{{dataset_type}}</td>
                            </tr>
                            <tr>
                                <td><b>Dataset size</b></td>
                                <td>{{dataset_size}}</td>
                            </tr>
                        </table>
                        </div>
                        <h4>Dataset labels</h4>
                        {{#show_dataset_labels}}
                        <p>The following label(s) are available for this dataset, and may for example be used for classification:</p>

                        <div class="table-container">
                            <table>
                                <thead>
                                <tr>
                                    <th><b>Label name</b></th>
                                    <th><b>Label classes</b></th>
                                </tr>
                                </thead>
                                <tbody>
                                {{#dataset_labels}}
                                <tr>
                                    <td>{{dataset_label_name}}</td>
                                    <td>{{dataset_label_classes}}</td>
                                </tr>
                                {{/dataset_labels}}
                                </tbody>
                            </table>
                        </div>
                        {{/show_dataset_labels}}
                        {{^show_dataset_labels}}
                        <p>This dataset does not have any available labels.</p>
                        {{/show_dataset_labels}}

                        {{#show_preprocessing}}
                        <h4>Preprocessings</h4>
                        The following preprocessing steps have been applied to the dataset:
                        <div>
                            <ul>
                                {{#preprocessing_sequence}}
                                <li>{{preprocessing_name}}; {{preprocessing_params}}</li>
                                {{/preprocessing_sequence}}
                            </ul>
                        </div>
                        {{/show_preprocessing}}
                        <h4>Download formats</h4>
                        <div>
                            <ul>
                            {{#formats}}
                                <li><a href="{{dataset_download_link}}" download>{{format_name}}</a></li>
                            {{/formats}}
                            </ul>
                        </div>
                    </div>
                </div>
            </div>
            {{/datasets}}
        </div>
    </div>
</body>
</html>